import unittest

import numpy as np
from torcheeg.transforms import ToGrid, ToInterpolatedGrid, To2d, MeanStdNormalize, MinMaxNormalize, BandSignal, BandDifferentialEntropy, BandPowerSpectralDensity, BandMeanAbsoluteDeviation, BandKurtosis, BandSkewness, Concatenate, MapChunk, PickElectrode, CWTSpectrum, ARRCoefficient, PearsonCorrelation, PhaseLockingCorrelation, BandApproximateEntropy, BandSampleEntropy, BandSVDEntropy, BandDetrendedFluctuationAnalysis, BandHiguchiFractalDimension, BandHjorth, BandHurst, BandPetrosianFractalDimension, BandBinPower, BandSpectralEntropy, DWTDecomposition, Downsample, Compose, RearrangeElectrode, Flatten
from torcheeg.datasets.constants import DEAP_CHANNEL_LOCATION_DICT, DEAP_CHANNEL_LIST, M3CV_CHANNEL_LOCATION_DICT


class TestNumpyTransforms(unittest.TestCase):
    def test_flatten(self):
        eeg = np.random.randn(62, 5)
        transformed_eeg = Flatten()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (310,))

    def test_dwt_decomposition(self):
        eeg = np.random.randn(32, 1000)
        transformed_eeg = DWTDecomposition()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (2, 32, 500))

    def test_cwt_spectrum(self):
        eeg = np.random.randn(32, 1000)
        transformed_eeg = CWTSpectrum()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 128, 1000))

        transformed_eeg = CWTSpectrum(contourf=True)(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 480, 640, 4))

    def test_pick_electrode(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = PickElectrode([1, 2])(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (2, 128))

        pick = [
            'FP1', 'AF3', 'F3', 'F7', 'FC5', 'FC1', 'C3', 'T7', 'CP5', 'CP1',
            'P3', 'P7', 'PO3', 'O1', 'FP2', 'AF4', 'F4', 'F8', 'FC6', 'FC2',
            'C4', 'T8', 'CP6', 'CP2', 'P4', 'P8', 'PO4', 'O2'
        ]
        pick_list = PickElectrode.to_index_list(pick, DEAP_CHANNEL_LIST)
        self.assertEqual(len(pick_list), 28)

        transformed_eeg = PickElectrode(
            PickElectrode.to_index_list(pick, DEAP_CHANNEL_LIST))(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (28, 128))

    def test_to_2d(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = To2d()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (1, 32, 128))

    def test_to_grid(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = ToGrid(DEAP_CHANNEL_LOCATION_DICT)(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (128, 9, 9))

    def test_to_interpolated_grid(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = ToInterpolatedGrid(DEAP_CHANNEL_LOCATION_DICT)(
            eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (128, 9, 9))

        eeg = np.random.randn(64, 128)
        transformed_eeg = ToInterpolatedGrid(M3CV_CHANNEL_LOCATION_DICT)(
            eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (128, 9, 11))

    def test_mean_std_normalize(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = MeanStdNormalize(axis=None)(eeg=eeg)
        self.assertEqual(eeg.shape, transformed_eeg['eeg'].shape)

        eeg = np.random.randn(32, 128)
        transformed_eeg = MeanStdNormalize(axis=0)(eeg=eeg)
        self.assertEqual(eeg.shape, transformed_eeg['eeg'].shape)

        eeg = np.random.randn(32, 128)
        transformed_eeg = MeanStdNormalize(axis=1)(eeg=eeg)
        self.assertEqual(eeg.shape, transformed_eeg['eeg'].shape)

    def test_min_max_normalize(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = MinMaxNormalize(axis=None)(eeg=eeg)
        self.assertEqual(eeg.shape, transformed_eeg['eeg'].shape)

        eeg = np.random.randn(32, 128)
        transformed_eeg = MinMaxNormalize(axis=0)(eeg=eeg)
        self.assertEqual(eeg.shape, transformed_eeg['eeg'].shape)

        eeg = np.random.randn(32, 128)
        transformed_eeg = MinMaxNormalize(axis=1)(eeg=eeg)
        self.assertEqual(eeg.shape, transformed_eeg['eeg'].shape)

    def test_band_signal(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandSignal()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (4, 32, 128))

    def test_band_differential_entropy(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandDifferentialEntropy()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_band_power_spectral_density(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandPowerSpectralDensity()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_band_mean_absolute_deviation(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandMeanAbsoluteDeviation()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_band_kurtosis(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandKurtosis()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_band_skewness(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandSkewness()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_arr_coefficient(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = ARRCoefficient(order=4)(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_concat(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = Concatenate([BandSkewness(), BandBinPower()])(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 8))

        transformed_eeg = Concatenate(
            [Compose([MinMaxNormalize(), BandSkewness()]),
             BandBinPower()])(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 8))

    def test_map_chunk(self):
        eeg = np.random.randn(64, 1000)
        transformed_eeg = MapChunk(BandDifferentialEntropy(),
                                   chunk_size=250,
                                   overlap=0)(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (64, 16))

        transformed_eeg = Concatenate([
            MapChunk(BandDifferentialEntropy(), chunk_size=250, overlap=0),
            MapChunk(BandDifferentialEntropy(), chunk_size=500, overlap=0),
            BandDifferentialEntropy()
        ])(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (64, 28))

        transformed_eeg = Concatenate([
            MapChunk(BandDifferentialEntropy(), chunk_size=250, overlap=0),
            MapChunk(BandDifferentialEntropy(), chunk_size=500, overlap=0),
            BandDifferentialEntropy(),
            BandPowerSpectralDensity()
        ])(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (64, 32))

    def test_pearson_correlation(self):
        eeg = np.random.randn(32, 4)
        transformed_eeg = PearsonCorrelation()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (1, 32, 32))

    def test_phase_locking_correlation(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = PhaseLockingCorrelation()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (1, 32, 32))

    def test_ap_entropy(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandApproximateEntropy()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_samp_entropy(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandSampleEntropy()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_svd_entropy(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandSVDEntropy()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_dfa(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandDetrendedFluctuationAnalysis()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_hfd(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandHiguchiFractalDimension()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_hjorth(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandHjorth()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_hurst(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandHurst()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_pdf(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandPetrosianFractalDimension()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_bin_power(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandBinPower()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 4))

    def test_spectral_entropy(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = BandSpectralEntropy()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 1))

    def test_downsample(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = Downsample(num_points=32, axis=-1)(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 32))

    def test_rearrange_electrode(self):
        src_eeg = np.random.rand(3, 128)
        tgt_eeg = RearrangeElectrode(source=['FP1', 'F3', 'F7'],
                                 target=['F3', 'F7', 'FP1', 'AF2'],
                                 missing='mean')(eeg=src_eeg)['eeg']
        self.assertEqual(tgt_eeg.shape, (4, 128))

        self.assertEqual(src_eeg[0, :].tolist(), tgt_eeg[2, :].tolist())
        self.assertEqual(src_eeg[1, :].tolist(), tgt_eeg[0, :].tolist())
        self.assertEqual(src_eeg[2, :].tolist(), tgt_eeg[1, :].tolist())


if __name__ == '__main__':
    unittest.main()


__version__ = '1.1.0'

import logging

from mne import set_log_level as _set_mne_log_level


def _set_lightning_log_level(level: str = "INFO"):

    # name to level
    level = getattr(logging, level)

    # configure logging at the root level of Lightning
    logging.getLogger("lightning.pytorch").setLevel(level)


def set_log_level(level: str = "DEBUG", third_party: str = "ERROR"):
    """
    Set log level for torcheeg.

    Args:
        level (str): Log level. Choose one of ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"].
        third_party (str): Log level for third party libraries. Choose one of ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"].
    """
    VALID_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    level = level.upper()
    if level not in VALID_LEVELS:
        raise ValueError(
            f"Invalid level {level}. Choose one of {VALID_LEVELS}.")

    _set_mne_log_level(third_party)
    _set_lightning_log_level(third_party)

    logging.basicConfig(
        level=level,
        format="%(asctime)s %(levelname)s %(threadName)s %(name)s %(message)s",
    )


set_log_level()


from typing import Union, Dict

import numpy as np
from spectrum import aryule

from ..base_transform import EEGTransform


class ARRCoefficient(EEGTransform):
    r'''
    Calculate autoregression reflection coefficients on the input data.

    .. code-block:: python

        transform = ARRCoefficient(order=4)
        transform(eeg=np.random.randn(32, 128))['eeg'].shape
        >>> (32, 4)

    Args:
        order (int): The order of autoregressive process to be fitted. (default: :obj:`4`)
        norm (str): Use a biased or unbiased correlation. (default: :obj:`biased`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)
    
    .. automethod:: __call__
    '''
    def __init__(self,
                 order: int = 4,
                 norm: str = 'biased',
                 apply_to_baseline: bool = False):
        super(ARRCoefficient,
              self).__init__(apply_to_baseline=apply_to_baseline)
        self.order = order
        self.norm = norm

    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals or features.
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            np.ndarray [number of electrodes, order]: The autoregression reflection coefficients.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def apply(self, eeg: np.ndarray, **kwargs):
        c_list = []
        for c in eeg:
            ar_coeffs, _, _ = aryule(c, order=self.order, norm=self.norm)
            c_list.append(ar_coeffs)
        return np.array(c_list)

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{
            'order': self.order,
            'norm': self.norm
        })


import unittest

import torch

from torcheeg.models import SimpleViT, ArjunViT, VanillaTransformer, ViT, Conformer


class TestTransformer(unittest.TestCase):

    def test_simple_vit(self):
        eeg = torch.randn(1, 128, 9, 9)
        model = SimpleViT(chunk_size=128, t_patch_size=32, s_patch_size=(3, 3), num_classes=2)
        pred = model(eeg)
        self.assertEqual(tuple(pred.shape), (1, 2))

        model = SimpleViT(chunk_size=128, t_patch_size=32, s_patch_size=(3, 3), num_classes=2)
        eeg = eeg.cuda()
        model = model.cuda()
        pred = model(eeg).cpu()
        self.assertEqual(tuple(pred.shape), (1, 2))

    def test_arjun_vit(self):
        eeg = torch.randn(1, 32, 128)
        model = ArjunViT(chunk_size=128, t_patch_size=32, num_classes=2)
        pred = model(eeg)
        self.assertEqual(tuple(pred.shape), (1, 2))

        eeg = eeg.cuda()
        model = model.cuda()
        pred = model(eeg).cpu()
        self.assertEqual(tuple(pred.shape), (1, 2))

    def test_vanilla_transformer(self):
        eeg = torch.randn(1, 32, 128)
        model = VanillaTransformer(chunk_size=128,
                                   t_patch_size=32,
                                   hid_channels=32,
                                   depth=3,
                                   heads=4,
                                   head_channels=64,
                                   mlp_channels=64,
                                   num_classes=2)
        pred = model(eeg)
        self.assertEqual(tuple(pred.shape), (1, 2))

        eeg = eeg.cuda()
        model = model.cuda()
        pred = model(eeg).cpu()
        self.assertEqual(tuple(pred.shape), (1, 2))

    def test_vit(self):
        eeg = torch.randn(1, 128, 9, 9)
        model = ViT(chunk_size=128, t_patch_size=32, s_patch_size=(3, 3), num_classes=2)
        pred = model(eeg)
        self.assertEqual(tuple(pred.shape), (1, 2))

        eeg = eeg.cuda()
        model = model.cuda()
        pred = model(eeg).cpu()
        self.assertEqual(tuple(pred.shape), (1, 2))

    def test_conformer(self):
        eeg = torch.randn(2, 1, 32, 128)
        model = Conformer(num_electrodes=32, sampling_rate=128, num_classes=2)
        pred = model(eeg)
        self.assertEqual(tuple(pred.shape), (2, 2))

        eeg = eeg.cuda()
        model = model.cuda()
        pred = model(eeg).cpu()
        self.assertEqual(tuple(pred.shape), (2, 2))


if __name__ == '__main__':
    unittest.main()



import torch.nn as nn
import math
import torch

import torch.nn.functional as F
from einops import rearrange
from einops.layers.torch import Rearrange, Reduce


class PatchEmbedding(nn.Module):
    def __init__(self,
                 num_electrodes: int,
                 hid_channels: int = 40,
                 dropout: float = 0.5):
        # self.patch_size = patch_size
        super().__init__()

        self.shallownet = nn.Sequential(
            nn.Conv2d(1, hid_channels, (1, 25), (1, 1)),
            nn.Conv2d(hid_channels, hid_channels, (num_electrodes, 1), (1, 1)),
            nn.BatchNorm2d(hid_channels),
            nn.ELU(),
            nn.AvgPool2d(
                (1, 75), (1, 15)
            ),  # pooling acts as slicing to obtain 'patch' along the time dimension as in ViT
            nn.Dropout(dropout),
        )

        self.projection = nn.Sequential(
            nn.Conv2d(hid_channels, hid_channels, (1, 1), stride=(
                1, 1)),  # transpose, conv could enhance fiting ability slightly
            Rearrange('b e (h) (w) -> b (h w) e'),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, _, _, _ = x.shape
        x = self.shallownet(x)
        x = self.projection(x)
        return x


class MultiHeadAttention(nn.Module):
    def __init__(self, hid_channels: int, heads: int, dropout: float):
        super().__init__()
        self.hid_channels = hid_channels
        self.heads = heads
        self.keys = nn.Linear(hid_channels, hid_channels)
        self.queries = nn.Linear(hid_channels, hid_channels)
        self.values = nn.Linear(hid_channels, hid_channels)
        self.att_drop = nn.Dropout(dropout)
        self.projection = nn.Linear(hid_channels, hid_channels)

    def forward(self,
                x: torch.Tensor,
                mask: torch.Tensor = None) -> torch.Tensor:
        queries = rearrange(self.queries(x),
                            "b n (h d) -> b h n d",
                            h=self.heads)
        keys = rearrange(self.keys(x), "b n (h d) -> b h n d", h=self.heads)
        values = rearrange(self.values(x), "b n (h d) -> b h n d", h=self.heads)
        energy = torch.einsum('bhqd, bhkd -> bhqk', queries, keys)
        if mask is not None:
            fill_value = torch.finfo(torch.float32).min
            energy.mask_fill(~mask, fill_value)

        scaling = self.hid_channels**(1 / 2)
        att = F.softmax(energy / scaling, dim=-1)
        att = self.att_drop(att)
        out = torch.einsum('bhal, bhlv -> bhav ', att, values)
        out = rearrange(out, "b h n d -> b n (h d)")
        out = self.projection(out)
        return out


class ResidualAdd(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        res = x
        x = self.fn(x, **kwargs)
        x += res
        return x


class FeedForwardBlock(nn.Sequential):
    def __init__(self,
                 hid_channels: int,
                 expansion: int = 4,
                 dropout: float = 0.):
        super().__init__(
            nn.Linear(hid_channels, expansion * hid_channels),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(expansion * hid_channels, hid_channels),
        )


class GELU(nn.Module):
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return input * 0.5 * (1.0 + torch.erf(input / math.sqrt(2.0)))


class TransformerEncoderBlock(nn.Sequential):
    def __init__(self, hid_channels: int, heads: int, dropout: float,
                 forward_expansion: int, forward_dropout: float):
        super().__init__(
            ResidualAdd(
                nn.Sequential(nn.LayerNorm(hid_channels),
                              MultiHeadAttention(hid_channels, heads, dropout),
                              nn.Dropout(dropout))),
            ResidualAdd(
                nn.Sequential(
                    nn.LayerNorm(hid_channels),
                    FeedForwardBlock(hid_channels,
                                     expansion=forward_expansion,
                                     dropout=forward_dropout),
                    nn.Dropout(dropout))))


class TransformerEncoder(nn.Sequential):
    def __init__(self,
                 depth: int,
                 hid_channels: int,
                 heads: int = 10,
                 dropout: float = 0.5,
                 forward_expansion: int = 4,
                 forward_dropout: float = 0.5):
        super().__init__(*[
            TransformerEncoderBlock(hid_channels=hid_channels,
                                    heads=heads,
                                    dropout=dropout,
                                    forward_expansion=forward_expansion,
                                    forward_dropout=forward_dropout)
            for _ in range(depth)
        ])


class ClassificationHead(nn.Sequential):
    def __init__(self,
                 in_channels: int,
                 num_classes: int,
                 hid_channels: int = 32,
                 dropout: float = 0.5):
        super().__init__()
        self.fc = nn.Sequential(nn.Linear(in_channels, hid_channels * 8),
                                nn.ELU(), nn.Dropout(dropout),
                                nn.Linear(hid_channels * 8, hid_channels),
                                nn.ELU(), nn.Dropout(dropout),
                                nn.Linear(hid_channels, num_classes))

    def forward(self, x):
        x = x.contiguous().view(x.size(0), -1)
        x = self.fc(x)
        return x


class Conformer(nn.Module):
    r'''
    The EEG Conformer model is based on the paper "EEG Conformer: Convolutional Transformer for EEG Decoding and Visualization". For more details, please refer to the following information. 

    It is worth noting that this model is not designed for EEG analysis, but shows good performance and can serve as a good research start.

    - Paper: Song Y, Zheng Q, Liu B, et al. EEG Conformer: Convolutional Transformer for EEG Decoding and Visualization[J]. IEEE Transactions on Neural Systems and Rehabilitation Engineering, 2022.
    - URL: https://ieeexplore.ieee.org/document/9991178
    - Related Project: https://github.com/eeyhsong/EEG-Conformer

    Below is a recommended suite for use in emotion recognition tasks:

    .. code-block:: python

        dataset = SEEDDataset(io_path=f'./seed',
                              root_path='./Preprocessed_EEG',
                              offline_transform=transforms.Compose([
                                  transforms.MinMaxNormalize(axis=-1),
                                  transforms.To2d()
                              ]),
                              online_transform=transforms.ToTensor(),
                              label_transform=transforms.Compose([
                                  transforms.Select('emotion'),
                                  transforms.Lambda(lambda x: x + 1)
                              ]))
        model = Conformer(num_electrodes=62,
                          sampling_rate=200,
                          hid_channels=40,
                          depth=6,
                          heads=10,
                          dropout=0.5,
                          forward_expansion=4,
                          forward_dropout=0.5,
                          num_classes=2)

    Args:
        num_electrodes (int): The number of electrodes. (default: :obj:`62`)
        sampling_rate (int): The sampling rate of EEG signals. (default: :obj:`200`)
        hid_channels (int): The feature dimension of embeded patch. (default: :obj:`40`)
        depth (int): The number of attention layers for each transformer block. (default: :obj:`6`)
        heads (int): The number of attention heads for each attention layer. (default: :obj:`10`)
        dropout (float): The dropout rate of the attention layer. (default: :obj:`0.5`)
        forward_expansion (int): The expansion factor of the feedforward layer. (default: :obj:`4`)
        forward_dropout (float): The dropout rate of the feedforward layer. (default: :obj:`0.5`)
        num_classes (int): The number of classes. (default: :obj:`2`)
    '''
    def __init__(self,
                 num_electrodes: int = 62,
                 sampling_rate: int = 200,
                 embed_dropout: float = 0.5,
                 hid_channels: int = 40,
                 depth: int = 6,
                 heads: int = 10,
                 dropout: float = 0.5,
                 forward_expansion: int = 4,
                 forward_dropout: float = 0.5,
                 cls_channels: int = 32,
                 cls_dropout: float = 0.5,
                 num_classes: int = 2):
        super().__init__()
        self.num_electrodes = num_electrodes
        self.sampling_rate = sampling_rate
        self.embed_dropout = embed_dropout
        self.hid_channels = hid_channels
        self.depth = depth
        self.heads = heads
        self.dropout = dropout
        self.forward_expansion = forward_expansion
        self.forward_dropout = forward_dropout
        self.cls_channels = cls_channels
        self.cls_dropout = cls_dropout
        self.num_classes = num_classes

        self.embd = PatchEmbedding(num_electrodes, hid_channels, embed_dropout)
        self.encoder = TransformerEncoder(depth,
                                          hid_channels,
                                          heads=heads,
                                          dropout=dropout,
                                          forward_expansion=forward_expansion,
                                          forward_dropout=forward_dropout)
        self.cls = ClassificationHead(in_channels=self.feature_dim(),
                                      num_classes=num_classes,
                                      hid_channels=cls_channels,
                                      dropout=cls_dropout)

    def feature_dim(self):
        with torch.no_grad():
            mock_eeg = torch.zeros(1, 1, self.num_electrodes,
                                   self.sampling_rate)

            mock_eeg = self.embd(mock_eeg)
            mock_eeg = self.encoder(mock_eeg)
            mock_eeg = mock_eeg.flatten(start_dim=1)

            return mock_eeg.shape[1]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.embd(x)
        x = self.encoder(x)
        x = x.flatten(start_dim=1)
        x = self.cls(x)
        return x


from typing import Dict, List, Union

from ..base_transform import LabelTransform


class FixCategory(LabelTransform):
    r'''
    Returns a pre-set label for all samples, usually used to supplement the dataset with new categories.

    .. code-block:: python

        transform = FixCategory(value=0)
        transform(y=3)['y']
        >>> 0

    :obj:`FixCategory` allows multiple values to be selected and returned as a list. Suitable for multi-classification tasks or multi-task learning.

    .. code-block:: python

        transform = FixCategory(value=[0, 1])
        transform(y=[1, 2])['y']
        >>> [0, 1]

    Args:
        value (str or list): The pre-set label.

    .. automethod:: __call__
    '''
    def __init__(self, value: Union[int, str, List]):
        super(FixCategory, self).__init__()
        self.value = value

    def __call__(self, *args, y: Dict, **kwargs) -> Union[int, float, List]:
        r'''
        Args:
            y (any): A label or label list describing the EEG signal samples.
            
        Returns:
            any: FixCategoryeded value pre-set by `value`.
        '''
        return super().__call__(*args, y=y, **kwargs)

    def apply(self, y: Dict, **kwargs) -> Union[int, float, List]:
        return self.value

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{
            'value': self.value
        })


import pdb
from typing import Dict, List, Callable

from ..base_transform import BaseTransform


class PDB(BaseTransform):
    r'''
    For debugging, insert breakpoints in transforms. The transformation itself does not change the input data.

    .. code-block:: python

        transform = Compose([
            ToTensor(),
            Resize(size=(64, 64)),
            PDB(),
            RandomNoise(p=0.1),
            RandomMask(p=0.1)
        ])
        transform(eeg=torch.randn(128, 9, 9))['eeg'].shape
        >>> (128, 64, 64)
    
    Args:
        targets (list): What data to transform via the PDB. (default: :obj:`['eeg', 'baseline', 'y']`)

    .. automethod:: __call__
    '''
    def __init__(self, targets: List[str] = ['eeg', 'baseline', 'y']):
        super(PDB, self).__init__()
        self._targets = targets

    def apply(self, *args, **kwargs) -> any:
        r'''
        Args:
            x (any): The input.

        Returns:
            any: The input.
        '''
        return args[0]

    def __call__(self, *args, **kwargs) -> Dict[str, any]:
        r'''
        Args:
            x (any): The input.
        Returns:
            any: The input.
        '''
        pdb.set_trace()
        return super().__call__(*args, **kwargs)

    @property
    def targets(self) -> Dict[str, Callable]:
        return {target: self.apply for target in self._targets}

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{'targets': [...]})



from typing import Tuple

import torch
import torch.nn as nn


class BGenerator(nn.Module):
    r'''
    TorchEEG provides an EEG feature generator based on CNN architecture and GAN for generating EEG grid representations of different frequency bands based on a given class label.

    .. code-block:: python

        g_model = BGenerator(in_channels=128)
        z = torch.normal(mean=0, std=1, size=(1, 128))
        fake_X = g_model(z)

    Args:
        in_channels (int): The input feature dimension (of noise vectors). (default: :obj:`128`)
        out_channels (int): The generated feature dimension of each electrode. (default: :obj:`4`)
        grid_size (tuple): Spatial dimensions of grid-like EEG representation. (default: :obj:`(9, 9)`)
    '''
    def __init__(self,
                 in_channels: int = 128,
                 out_channels: int = 4,
                 grid_size: Tuple[int, int] = (9, 9)):
        super(BGenerator, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.grid_size = grid_size

        self.deproj = nn.Sequential(
            nn.Linear(in_channels, in_channels * 4 * 3 * 3), nn.LeakyReLU())

        self.deconv1 = nn.Sequential(
            nn.ConvTranspose2d(in_channels * 4,
                               in_channels * 2,
                               kernel_size=3,
                               stride=2,
                               padding=1,
                               bias=True), nn.BatchNorm2d(in_channels * 2),
            nn.LeakyReLU())
        self.deconv2 = nn.Sequential(
            nn.ConvTranspose2d(in_channels * 2,
                               in_channels * 2,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=True), nn.BatchNorm2d(in_channels * 2),
            nn.LeakyReLU())
        self.deconv3 = nn.Sequential(
            nn.ConvTranspose2d(in_channels * 2,
                               in_channels,
                               kernel_size=3,
                               stride=2,
                               padding=1,
                               bias=True), nn.BatchNorm2d(in_channels),
            nn.LeakyReLU())
        self.deconv4 = nn.ConvTranspose2d(in_channels,
                                          out_channels,
                                          kernel_size=3,
                                          stride=1,
                                          padding=1,
                                          bias=True)

    def forward(self, x: torch.Tensor):
        r'''
        Args:
            x (torch.Tensor): a random vector, the ideal input shape is :obj:`[n, 128]`. Here, :obj:`n` corresponds to the batch size, and :obj:`128` corresponds to :obj:`in_channels`.

        Returns:
            torch.Tensor[n, 4, 9, 9]: the generated fake EEG signals. Here, :obj:`4` corresponds to the :obj:`out_channels`, and :obj:`(9, 9)` corresponds to the :obj:`grid_size`.
        '''
        x = self.deproj(x)
        x = x.view(-1, self.in_channels * 4, 3, 3)
        x = self.deconv1(x)
        x = self.deconv2(x)
        x = self.deconv3(x)
        x = self.deconv4(x)
        return x


class BDiscriminator(nn.Module):
    r'''
    TorchEEG provides an EEG feature generator based on CNN architecture and GAN for generating EEG grid representations of different frequency bands based on a given class label.

    .. code-block:: python

        g_model = BGenerator(in_channels=128)
        d_model = BDiscriminator(in_channels=4)
        z = torch.normal(mean=0, std=1, size=(1, 128))
        fake_X = g_model(z)
        disc_X = d_model(fake_X)

    Args:
        in_channels (int): The feature dimension of each electrode. (default: :obj:`4`)
        grid_size (tuple): Spatial dimensions of grid-like EEG representation. (default: :obj:`(9, 9)`)
        hid_channels (int): The number of hidden nodes in the first fully connected layer. (default: :obj:`32`)
    '''
    def __init__(self,
                 in_channels: int = 4,
                 grid_size: Tuple[int, int] = (9, 9),
                 hid_channels: int = 64):
        super(BDiscriminator, self).__init__()

        self.in_channels = in_channels
        self.grid_size = grid_size
        self.hid_channels = hid_channels

        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels,
                      hid_channels,
                      kernel_size=3,
                      stride=1,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels), nn.LeakyReLU())
        self.conv2 = nn.Sequential(
            nn.Conv2d(hid_channels,
                      hid_channels * 2,
                      kernel_size=3,
                      stride=2,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels * 2),
            nn.LeakyReLU())
        self.conv3 = nn.Sequential(
            nn.Conv2d(hid_channels * 2,
                      hid_channels * 2,
                      kernel_size=3,
                      stride=1,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels * 2),
            nn.LeakyReLU())
        self.conv4 = nn.Sequential(
            nn.Conv2d(hid_channels * 2,
                      hid_channels * 4,
                      kernel_size=3,
                      stride=2,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels * 4),
            nn.LeakyReLU())

        self.proj = nn.Linear(self.feature_dim, 1)

    @property
    def feature_dim(self):
        with torch.no_grad():
            mock_eeg = torch.zeros(1, self.in_channels, *self.grid_size)

            mock_eeg = self.conv1(mock_eeg)
            mock_eeg = self.conv2(mock_eeg)
            mock_eeg = self.conv3(mock_eeg)
            mock_eeg = self.conv4(mock_eeg)

        return mock_eeg.flatten(start_dim=1).shape[-1]

    def forward(self, x: torch.Tensor):
        r'''
        Args:
            x (torch.Tensor): EEG signal representation, the ideal input shape is :obj:`[n, 4, 9, 9]`. Here, :obj:`n` corresponds to the batch size, :obj:`4` corresponds to the :obj:`in_channels`, and :obj:`(9, 9)` corresponds to the :obj:`grid_size`.

        Returns:
            torch.Tensor[n, 1]: Predicts the result of whether a given sample is a fake sample or not. Here, :obj:`n` corresponds to the batch size.
        '''
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = x.flatten(start_dim=1)
        x = self.proj(x)
        return x


from pathlib import Path
from typing import Any, Callable, Dict, Tuple, Union

import mne
import numpy as np

from .base_dataset import BaseDataset


def default_read_fn(file_path, **kwargs):
    # Load EEG file
    raw = mne.io.read_raw(file_path)
    # Convert raw to epochs
    epochs = mne.make_fixed_length_epochs(raw, duration=1)
    # Return EEG data
    return epochs


class FolderDataset(BaseDataset):
    '''
    Read EEG samples and their corresponding labels from a fixed folder structure. This class allows two kinds of common file structures, :obj:`subject_in_label` and :obj:`label_in_subject`. Here, :obj:`subject_in_label` corresponds to the following file structure:

    .. code-block:: python

        tree
        # outputs
        label01
        |- sub01.edf
        |- sub02.edf
        label02
        |- sub01.edf
        |- sub02.edf

    And :obj:`label_in_subject` corresponds to the following file structure:

    .. code-block:: python

        tree
        # outputs
        sub01
        |- label01.edf
        |- label02.edf
        sub02
        |- label01.edf
        |- label02.edf

    An example dataset for GNN-based methods:

    .. code-block:: python
    
        sfreq = 128  # Sampling rate
        n_channels = 14  # Number of channels
        duration = 5  # Data collected for 5 seconds
        for i in range(num_files):
            n_samples = sfreq * duration
            data = np.random.randn(n_channels, n_samples)

            ch_names = [f'ch_{i+1:03}' for i in range(n_channels)]
            ch_types = ['eeg'] * n_channels
            info = mne.create_info(ch_names, sfreq, ch_types)
            raw = mne.io.RawArray(data, info)

            file_name = f'sub{i+1}.fif'
            file_path = os.path.join('./root_folder/', file_name)
            raw.save(file_path)

        label_map = {'folder1': 0, 'folder2': 1}
        dataset = FolderDataset(io_path='./folder',
                                root_path='./root_folder',
                                structure='subject_in_label',
                                num_channel=14,
                                online_transform=transforms.ToTensor(),
                                label_transform=transforms.Compose([
                                    transforms.Select('label'),
                                    transforms.Lambda(lambda x: label_map[x])
                                ]),
                                num_worker=4)

    Args:
        root_path (str): The path to the root folder. (default: :obj:`'./folder'`)
        structure (str): Folder structure, which affects how labels and subjects are mapped to EEG signal samples. Please refer to the above description of the structure of the two folders to select the correct parameters. (default: :obj:`'subject_in_label'`)
        read_fn (Callable): Method for reading files in a folder. By default, this class provides methods for reading files using :obj:`mne.io.read_raw`. At the same time, we allow users to pass in custom file reading methods. The first input parameter of whose is file_path, and other parameters are additional parameters passed in when the class is initialized. For example, you can pass :obj:`chunk_size=32` to :obj:`FolderDataset`, then :obj:`chunk_size` will be received here.
        online_transform (Callable, optional): The transformation of the EEG signals and baseline EEG signals. The input is a :obj:`np.ndarray`, and the ouput is used as the first and second value of each element in the dataset. (default: :obj:`None`)
        offline_transform (Callable, optional): The usage is the same as :obj:`online_transform`, but executed before generating IO intermediate results. (default: :obj:`None`)
        label_transform (Callable, optional): The transformation of the label. The input is an information dictionary, and the ouput is used as the third value of each element in the dataset. (default: :obj:`None`)
        io_path (str): The path to generated unified data IO, cached as an intermediate result. (default: :obj:`./io/deap`)
        io_size (int): Maximum size database may grow to; used to size the memory mapping. If database grows larger than ``map_size``, an exception will be raised and the user must close and reopen. (default: :obj:`10485760`)
        io_mode (str): Storage mode of EEG signal. When io_mode is set to :obj:`lmdb`, TorchEEG provides an efficient database (LMDB) for storing EEG signals. LMDB may not perform well on limited operating systems, where a file system based EEG signal storage is also provided. When io_mode is set to :obj:`pickle`, pickle-based persistence files are used. (default: :obj:`lmdb`)
        num_worker (int): Number of subprocesses to use for data loading. 0 means that the data will be loaded in the main process. (default: :obj:`0`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
        in_memory (bool): Whether to load the entire dataset into memory. If :obj:`in_memory` is set to True, then the first time an EEG sample is read, the entire dataset is loaded into memory for subsequent retrieval. Otherwise, the dataset is stored on disk to avoid the out-of-memory problem. (default: :obj:`False`)
    '''
    def __init__(self,
                 root_path: str = './folder',
                 structure: str = 'subject_in_label',
                 read_fn: Union[None, Callable] = default_read_fn,
                 online_transform: Union[None, Callable] = None,
                 offline_transform: Union[None, Callable] = None,
                 label_transform: Union[None, Callable] = None,
                 io_path: str = './io/folder',
                 io_size: int = 10485760,
                 io_mode: str = 'lmdb',
                 num_worker: int = 0,
                 verbose: bool = True,
                 in_memory: bool = False,
                 **kwargs):
        # pass all arguments to super class
        params = {
            'root_path': root_path,
            'structure': structure,
            'read_fn': read_fn,
            'online_transform': online_transform,
            'offline_transform': offline_transform,
            'label_transform': label_transform,
            'io_path': io_path,
            'io_size': io_size,
            'io_mode': io_mode,
            'num_worker': num_worker,
            'verbose': verbose,
            'in_memory': in_memory
        }
        params.update(kwargs)
        super().__init__(**params)
        # save all arguments to __dict__
        self.__dict__.update(params)

    @staticmethod
    def process_record(file: Any = None,
                   offline_transform: Union[None, Callable] = None,
                   read_fn: Union[None, Callable] = None,
                   **kwargs):

        file_path, subject_id, label = file

        trial_samples = read_fn(file_path, **kwargs)
        events = [i[0] for i in trial_samples.events]
        events.append(
            events[-1] +
            np.diff(events)[0])  # time interval between all events are same

        write_pointer = 0
        for i, trial_signal in enumerate(trial_samples.get_data()):
            t_eeg = trial_signal
            if not offline_transform is None:
                t = offline_transform(eeg=trial_signal)
                t_eeg = t['eeg']

            clip_id = f'{subject_id}_{label}_{write_pointer}'
            write_pointer += 1

            record_info = {
                'subject_id': subject_id,
                'start_at': events[i],
                'end_at': events[i + 1],
                'clip_id': clip_id,
                'label': label
            }

            yield {'eeg': t_eeg, 'key': clip_id, 'info': record_info}

    def set_records(self, root_path: str = './folder',
                   structure: str = 'subject_in_label',
                   **kwargs):
        # get all the subfolders
        subfolders = [str(i) for i in Path(root_path).iterdir() if i.is_dir()]
        # get all the files in the subfolders
        file_path_list = []
        for subfolder in subfolders:
            file_path_list += [
                str(i) for i in Path(subfolder).iterdir() if i.is_file()
            ]
        # get the subject id
        if structure == 'subject_in_label':
            # get the file name without the extension
            subjects = [i.split('/')[-1].split('.')[0] for i in file_path_list]
            labels = [i.split('/')[-2] for i in file_path_list]
        elif structure == 'label_in_subject':
            subjects = [i.split('/')[-2] for i in file_path_list]
            labels = [i.split('/')[-1].split('.')[0] for i in file_path_list]
        else:
            raise ValueError('Unknown folder mode: {}'.format(structure))

        file_path_subject_label = list(zip(file_path_list, subjects, labels))
        return file_path_subject_label

    def __getitem__(self, index: int) -> Tuple[any, any, int, int, int]:
        info = self.read_info(index)

        eeg_index = str(info['clip_id'])
        eeg_record = str(info['_record_id'])
        eeg = self.read_eeg(eeg_record, eeg_index)

        signal = eeg
        label = info

        if self.online_transform:
            signal = self.online_transform(eeg=eeg)['eeg']

        if self.label_transform:
            label = self.label_transform(y=info)['y']

        return signal, label

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'root_path': self.root_path,
                'chunk_size': self.chunk_size,
                'overlap': self.overlap,
                'num_channel': self.num_channel,
                'online_transform': self.online_transform,
                'offline_transform': self.offline_transform,
                'label_transform': self.label_transform,
                'before_trial': self.before_trial,
                'after_trial': self.after_trial,
                'num_worker': self.num_worker,
                'verbose': self.verbose,
                'io_size': self.io_size
            })



import os
import random
import shutil
import unittest

from torcheeg import transforms
from torcheeg.datasets import (AMIGOSDataset, DEAPDataset, DREAMERDataset,
                               MAHNOBDataset, SEEDDataset, SEEDFeatureDataset,
                               SEEDIVDataset, SEEDIVFeatureDataset,
                               MPEDFeatureDataset, BCI2022Dataset)


class TestEmotionRecognitionDataset(unittest.TestCase):
    def setUp(self):
        if os.path.exists('./tmp_out/'):
            shutil.rmtree('./tmp_out/')
        os.mkdir('./tmp_out/')

    def test_mped_feature_dataset(self):
        io_path = f'./tmp_out/mped_feature_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/EEG_feature'

        dataset = MPEDFeatureDataset(
            io_path=io_path,
            root_path=root_path,
            online_transform=transforms.ToTensor(),
            label_transform=transforms.Compose([
                transforms.Select('emotion'),
                transforms.Lambda(lambda x: int(x) + 1),
            ]),
            num_worker=4)

        self.assertEqual(len(dataset), 129904)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (62, 5))
        last_item = dataset[129903]
        self.assertEqual(last_item[0].shape, (62, 5))

    def test_mahnob_dataset(self):
        io_path = f'./tmp_out/mahnob_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/Sessions'

        dataset = MAHNOBDataset(io_path=io_path,
                                root_path=root_path,
                                online_transform=transforms.ToTensor(),
                                label_transform=transforms.Compose([
                                    transforms.Select('feltVlnc'),
                                    transforms.Binary(5.0),
                                ]),
                                num_worker=4)

        self.assertEqual(len(dataset), 16410)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (32, 128))
        last_item = dataset[16409]
        self.assertEqual(last_item[0].shape, (32, 128))

    def test_amigos_dataset(self):
        io_path = f'./tmp_out/amigos_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/data_preprocessed'

        dataset = AMIGOSDataset(io_path=io_path,
                                root_path=root_path,
                                num_trial=16,
                                online_transform=transforms.ToTensor(),
                                label_transform=transforms.Compose([
                                    transforms.Select('valence'),
                                    transforms.Binary(5.0),
                                ]),
                                num_worker=4)

        self.assertEqual(len(dataset), 45474)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (14, 128))
        last_item = dataset[45473]
        self.assertEqual(last_item[0].shape, (14, 128))

    def test_deap_dataset(self):
        io_path = f'./tmp_out/deap_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/data_preprocessed_python'

        dataset = DEAPDataset(io_path=io_path,
                              root_path=root_path,
                              online_transform=transforms.ToTensor(),
                              label_transform=transforms.Compose([
                                  transforms.Select('valence'),
                                  transforms.Binary(5.0),
                              ]),
                              num_worker=4)
        self.assertEqual(len(dataset), 76800)
        self.assertEqual(len(dataset.eeg_io), 78080)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (32, 128))
        last_item = dataset[76799]
        self.assertEqual(last_item[0].shape, (32, 128))

    def test_dreamer_dataset(self):
        io_path = f'./tmp_out/dreamer_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        mat_path = './tmp_in/DREAMER.mat'

        dataset = DREAMERDataset(io_path=io_path,
                                 mat_path=mat_path,
                                 online_transform=transforms.ToTensor(),
                                 label_transform=transforms.Compose([
                                     transforms.Select('valence'),
                                     transforms.Binary(3.0),
                                 ]),
                                 num_worker=4)

        self.assertEqual(len(dataset), 85744)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (14, 128))
        last_item = dataset[85743]
        self.assertEqual(last_item[0].shape, (14, 128))

    def test_seed_dataset(self):
        io_path = f'./tmp_out/seed_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/Preprocessed_EEG'

        dataset = SEEDDataset(io_path=io_path,
                              root_path=root_path,
                              online_transform=transforms.ToTensor(),
                              label_transform=transforms.Compose([
                                  transforms.Select('emotion'),
                                  transforms.Lambda(lambda x: int(x) + 1),
                              ]),
                              num_worker=4)

        self.assertEqual(len(dataset), 152730)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (62, 200))
        last_item = dataset[152729]
        self.assertEqual(last_item[0].shape, (62, 200))

    def test_seed_feature_dataset(self):
        io_path = f'./tmp_out/seed_feature_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/ExtractedFeatures'

        dataset = SEEDFeatureDataset(
            io_path=io_path,
            root_path=root_path,
            online_transform=transforms.ToTensor(),
            label_transform=transforms.Compose([
                transforms.Select('emotion'),
                transforms.Lambda(lambda x: int(x) + 1),
            ]),
            num_worker=4)

        self.assertEqual(len(dataset), 152730)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (62, 5))
        last_item = dataset[152729]
        self.assertEqual(last_item[0].shape, (62, 5))

    def test_seed_iv_dataset(self):
        io_path = f'./tmp_out/seed_iv_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/eeg_raw_data'

        dataset = SEEDIVDataset(io_path=io_path,
                                root_path=root_path,
                                online_transform=transforms.ToTensor(),
                                label_transform=transforms.Select('emotion'),
                                num_worker=4)

        self.assertEqual(len(dataset), 37575)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (62, 800))
        last_item = dataset[37574]
        self.assertEqual(last_item[0].shape, (62, 800))

    def test_seed_iv_feature_dataset(self):
        io_path = f'./tmp_out/seed_iv_feature_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/eeg_feature_smooth'

        dataset = SEEDIVFeatureDataset(
            io_path=io_path,
            root_path=root_path,
            online_transform=transforms.ToTensor(),
            label_transform=transforms.Select('emotion'),
            num_worker=4)

        self.assertEqual(len(dataset), 37575)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (62, 5))
        last_item = dataset[37574]
        self.assertEqual(last_item[0].shape, (62, 5))

    def test_bci2022(self):
        io_path = f'./tmp_out/bci2022_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/TrainSet'

        dataset = BCI2022Dataset(io_path=io_path,
                                 root_path=root_path,
                                 online_transform=transforms.ToTensor(),
                                 label_transform=transforms.Select('emotion'),
                                 channel_num=30,
                                 num_worker=4)

        self.assertEqual(len(dataset), 146812)
        first_item = dataset[0]
        self.assertEqual(first_item[0].shape, (30, 250))
        last_item = dataset[146811]
        self.assertEqual(last_item[0].shape, (30, 250))


if __name__ == '__main__':
    unittest.main()



from typing import Dict, List, Union

from ..base_transform import LabelTransform


class Mapping(LabelTransform):
    r'''
    Mapping the label according to a certain dictionary.
    
    .. code-block:: python

        transform = Mapping({
            'left_hand': 0,
            'right_hand': 1,
        })
        transform(y='left_hand')['y']
        >>> 0

    :obj:`Mapping` allows simultaneous binarization using the same threshold for multiple labels.

    .. code-block:: python

        transform = Mapping({
            'left_hand': 0,
            'right_hand': 1,
            'left_feet': 0,
            'right_feet': 1
        })
        transform(y=['left_hand', 'left_feet'])['y']
        >>> [0, 0]

    Args:
        map_dict (float): The mapping dictionary.
        default (float, optional): The default value when the input label is not in the dictionary. (default: :obj:`-1`)

    .. automethod:: __call__
    '''
    def __init__(self, map_dict: float, default: float = -1) -> None:
        super(Mapping, self).__init__()
        self.map_dict = map_dict
        self.default = default

    def __call__(self, *args, y: Union[int, float, List],
                 **kwargs) -> Union[int, List]:
        r'''
        Args:
            label (int, float, or list): The input label or list of labels.
            
        Returns:
            int, float, or list: The output label or list of labels after binarization.
        '''
        return super().__call__(*args, y=y, **kwargs)

    def apply(self, y: Union[int, float, List], **kwargs) -> Union[int, List]:
        if isinstance(y, list):
            return [self.map_dict.get(label, self.default) for label in y]
        else:
            return self.map_dict.get(y, self.default)

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{'map_dict': self.map_dict})


from typing import Dict, Union

import torch
import numpy as np
from sklearn.decomposition import PCA
from torch.fft import fft, ifft
from torch.nn.functional import pad

from ..base_transform import EEGTransform


class RandomEEGTransform(EEGTransform):
    def __init__(self, p: float = 0.5, apply_to_baseline: bool = False):
        super(RandomEEGTransform,
              self).__init__(apply_to_baseline=apply_to_baseline)
        self.p = p

    def apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        if self.p < torch.rand(1):
            return eeg
        return self.random_apply(eeg, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        raise NotImplementedError

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{'p': self.p})


class RandomNoise(RandomEEGTransform):
    '''
    Add random noise conforming to the normal distribution on the EEG signal.
    
    .. code-block:: python

        transform = RandomNoise(p=0.5)
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

    Args:
        mean (float): The mean of the normal distribution of noise. (default: :obj:`0.0`)
        std (float): The standard deviation of the normal distribution of noise. (default: :obj:`0.0`)
        p (float): Probability of adding noise to EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no noise is added to every sample and 1.0 means that noise is added to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 mean: float = 0.0,
                 std: float = 1.0,
                 p: float = 0.5,
                 apply_to_baseline: bool = False):
        super(RandomNoise, self).__init__(p=p,
                                          apply_to_baseline=apply_to_baseline)
        self.mean = mean
        self.std = std

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after adding random noise.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        noise = torch.randn_like(eeg)
        noise = (noise + self.mean) * self.std
        return eeg + noise

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{'mean': self.mean, 'std': self.std})


class RandomMask(RandomEEGTransform):
    '''
    Overlay the EEG signal using a random mask, and the value of the overlaid data points was set to 0.0.
    
    .. code-block:: python

        transform = RandomMask()
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

    Args:
        ratio (float): The proportion of data points covered by the mask out of all data points for each EEG signal sample. (default: :obj:`0.5`)
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 ratio: float = 0.5,
                 p: float = 0.5,
                 apply_to_baseline: bool = False):
        super(RandomMask, self).__init__(p=p,
                                         apply_to_baseline=apply_to_baseline)
        self.ratio = ratio

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random mask.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        mask = torch.rand_like(eeg)
        mask = (mask < self.ratio).to(eeg.dtype)
        return eeg * mask

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{'ratio': self.ratio})


class RandomMask2D(RandomEEGTransform):
    '''
    Overlay the EEG signal using a random mask, and the value of the overlaid data points was set to 0.0. The mask is applied to the three dimensions of the input tensor, where the first dimension represents the number of channels (bands, or something like that), the second dimension denotes the number of electrodes and the second dimension represents the number of time points. The mask is applied to the electrode dimension.
    
    .. code-block:: python

        transform = RandomMask()
        transform(eeg=torch.randn(1, 32, 128))['eeg'].shape
        >>> (1, 32, 128)

    Args:
        ratio (float): The proportion of data points covered by the mask out of all data points for each EEG signal sample. (default: :obj:`0.5`)
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 ratio: float = 0.5,
                 p: float = 0.5,
                 apply_to_baseline: bool = False):
        super(RandomMask2D, self).__init__(p=p,
                                           apply_to_baseline=apply_to_baseline)
        self.ratio = ratio

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random mask.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        # mask at the electrode dimension
        mask = torch.rand(eeg.shape[1], device=eeg.device)
        mask = (mask < self.ratio).to(eeg.dtype)
        # to the same shape of eeg
        mask = mask.unsqueeze(0).unsqueeze(2).repeat(eeg.shape[0], 1,
                                                     eeg.shape[2])
        return eeg * mask

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{'ratio': self.ratio})


class RandomMaskGrid(RandomEEGTransform):
    '''
    Overlay the EEG signal using a random mask, and the value of the overlaid data points was set to 0.0. The mask is applied to the last three dimensions of the input tensor, where the second dimension represents the height of the grid and the third dimension represents the width of the grid. The mask is applied to the height and width dimensions.
    
    .. code-block:: python

        transform = RandomMaskGrid()
        transform(eeg=torch.randn(4, 9, 9))['eeg'].shape
        >>> (4, 9, 9)

    Args:
        ratio (float): The proportion of data points covered by the mask out of all data points for each EEG signal sample. (default: :obj:`0.5`)
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 ratio: float = 0.5,
                 p: float = 0.5,
                 apply_to_baseline: bool = False):
        super(RandomMaskGrid,
              self).__init__(p=p, apply_to_baseline=apply_to_baseline)
        self.ratio = ratio

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random mask.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        # mask at the electrode dimension
        mask = torch.rand(eeg.shape[1], eeg.shape[2], device=eeg.device)
        mask = (mask < self.ratio).to(eeg.dtype)
        mask = mask.unsqueeze(0).repeat(eeg.shape[0], 1, 1)
        return eeg * mask

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{'ratio': self.ratio})


class RandomWindowSlice(RandomEEGTransform):
    '''
    Randomly applies a slice transformation with a given probability, where the original time series is sliced by a window, and the sliced data is scaled to the original size. It is worth noting that the random position where each channel slice starts is the same.
    
    .. code-block:: python

        transform = RandomWindowSlice()
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

        transform = RandomWindowSlice(window_size=100)
        transform(eeg=torch.randn(1, 32, 128))['eeg'].shape
        >>> (1, 32, 128)

        transform = RandomWindowSlice(p=1.0, series_dim=0)
        transform(eeg=torch.randn(128, 9, 9))['eeg'].shape
        >>> (128, 9, 9)

    Args:
        window_size (int): The window size of the slice, the original signal will be sliced to the window_size size, and then adaptively scaled to the input shape.
        series_dim (int): Dimension of the time series in the input tensor. (default: :obj:`-1`)
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 window_size: int = 120,
                 series_dim: int = -1,
                 p: float = 0.5,
                 apply_to_baseline: bool = False):
        super(RandomWindowSlice,
              self).__init__(p=p, apply_to_baseline=apply_to_baseline)
        self.window_size = window_size
        self.series_dim = series_dim

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random window slicing.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        eeg = eeg.numpy()

        assert len(eeg.shape) == 2 or len(
            eeg.shape
        ) == 3, 'Window slicing is only supported on 2D arrays or 3D arrays. In 2D arrays, the last dimension represents time series. In 3D arrays, the second dimension represents time series.'

        if self.window_size >= eeg.shape[self.series_dim]:
            return eeg

        assert -len(eeg.shape) <= self.series_dim < len(
            eeg.shape
        ), f'series_dim should be in range of [{- len(eeg.shape)}, {len(eeg.shape)}).'

        if self.series_dim < 0:
            self.series_dim = len(eeg.shape) + self.series_dim

        if self.series_dim != (len(eeg.shape) - 1):
            transpose_dims = list(range(len(eeg.shape)))
            transpose_dims.pop(self.series_dim)
            transpose_dims = [*transpose_dims, self.series_dim]
            eeg = eeg.transpose(transpose_dims)

        if len(eeg.shape) == 2:
            starts = np.random.randint(low=0,
                                       high=eeg.shape[-1] - self.window_size,
                                       size=(eeg.shape[0])).astype(int)
        else:
            starts = np.random.randint(low=0,
                                       high=eeg.shape[-1] - self.window_size,
                                       size=(eeg.shape[0],
                                             eeg.shape[1])).astype(int)
        ends = (self.window_size + starts).astype(int)

        new_eeg = np.zeros_like(eeg)
        for i, eeg_i in enumerate(eeg):
            if len(eeg.shape) == 3:
                for j, eeg_i_j in enumerate(eeg_i):
                    new_eeg[i][j] = np.interp(
                        np.linspace(0, self.window_size, num=eeg.shape[-1]),
                        np.arange(self.window_size),
                        eeg_i_j[starts[i][j]:ends[i][j]]).T
            else:
                new_eeg[i] = np.interp(
                    np.linspace(0, self.window_size, num=eeg.shape[-1]),
                    np.arange(self.window_size), eeg_i[starts[i]:ends[i]]).T

        if self.series_dim != (len(eeg.shape) - 1):
            undo_transpose_dims = [0] * len(eeg.shape)
            for i, dim in enumerate(transpose_dims):
                undo_transpose_dims[dim] = i
            new_eeg = new_eeg.transpose(undo_transpose_dims)

        return torch.from_numpy(new_eeg)

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'window_size': self.window_size,
                'series_dim': self.series_dim
            })


class RandomWindowWarp(RandomEEGTransform):
    '''
    Apply the window warping with a given probability, where a part of time series data is warpped by speeding it up or down.
    
    .. code-block:: python

        transform = RandomWindowWarp()
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

        transform = RandomWindowWarp(window_size=24, warp_size=48)
        transform(eeg=torch.randn(1, 32, 128))['eeg'].shape
        >>> (1, 32, 128)

        transform = RandomWindowWarp(p=1.0, series_dim=0)
        transform(eeg=torch.randn(128, 9, 9))['eeg'].shape
        >>> (128, 9, 9)

    Args:
        window_size (int): Randomly pick a window of size window_size on the time series to transform. (default: :obj:`-1`)
        warp_size (int): The size of the window after the warp. If warp_size is larger than window_size, it means slowing down, and if warp_size is smaller than window_size, it means speeding up. (default: :obj:`24`)
        series_dim (int): Dimension of the time series in the input tensor. (default: :obj:`-1`)
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 window_size: int = 12,
                 warp_size: int = 24,
                 series_dim: int = -1,
                 p: float = 0.5,
                 apply_to_baseline: bool = False):
        super(RandomWindowWarp,
              self).__init__(p=p, apply_to_baseline=apply_to_baseline)
        self.window_size = window_size
        self.warp_size = warp_size
        self.series_dim = series_dim

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random window warping.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        eeg = eeg.numpy()

        assert len(eeg.shape) == 2 or len(
            eeg.shape
        ) == 3, 'Window warping is only supported on 2D arrays or 3D arrays. In 2D arrays, the last dimension represents time series. In 3D arrays, the second dimension represents time series.'

        if self.window_size >= eeg.shape[self.series_dim]:
            return eeg

        if self.series_dim != (len(eeg.shape) - 1):
            transpose_dims = list(range(len(eeg.shape)))
            transpose_dims.pop(self.series_dim)
            transpose_dims = [*transpose_dims, self.series_dim]
            eeg = eeg.transpose(transpose_dims)

        window_steps = np.arange(self.window_size)
        if len(eeg.shape) == 2:
            starts = np.random.randint(low=0,
                                       high=eeg.shape[-1] - self.window_size,
                                       size=(eeg.shape[0])).astype(int)
        else:
            starts = np.random.randint(low=0,
                                       high=eeg.shape[-1] - self.window_size,
                                       size=(eeg.shape[0],
                                             eeg.shape[1])).astype(int)
        ends = (self.window_size + starts).astype(int)

        new_eeg = np.zeros_like(eeg)
        for i, eeg_i in enumerate(eeg):
            if len(eeg.shape) == 3:
                for j, eeg_i_j in enumerate(eeg_i):
                    start_seg = eeg_i_j[:starts[i][j]]
                    window_seg = np.interp(
                        np.linspace(0, self.window_size - 1,
                                    num=self.warp_size), window_steps,
                        eeg_i_j[starts[i][j]:ends[i][j]])
                    end_seg = eeg_i_j[ends[i][j]:]
                    warped = np.concatenate((start_seg, window_seg, end_seg))
                    new_eeg[i][j] = np.interp(
                        np.arange(eeg.shape[-1]),
                        np.linspace(0, eeg.shape[-1] - 1., num=warped.size),
                        warped).T
            else:
                start_seg = eeg_i[:starts[i]]
                window_seg = np.interp(
                    np.linspace(0, self.window_size - 1, num=self.warp_size),
                    window_steps, eeg_i[starts[i]:ends[i]])
                end_seg = eeg_i[ends[i]:]
                warped = np.concatenate((start_seg, window_seg, end_seg))
                new_eeg[i] = np.interp(
                    np.arange(eeg.shape[-1]),
                    np.linspace(0, eeg.shape[-1] - 1., num=warped.size),
                    warped).T

        if self.series_dim != (len(eeg.shape) - 1):
            undo_transpose_dims = [0] * len(eeg.shape)
            for i, dim in enumerate(transpose_dims):
                undo_transpose_dims[dim] = i
            new_eeg = new_eeg.transpose(undo_transpose_dims)

        return torch.from_numpy(new_eeg)

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'window_size': self.window_size,
                'warp_size': self.warp_size,
                'series_dim': self.series_dim
            })


class RandomPCANoise(RandomEEGTransform):
    '''
    Add noise with a given probability, where the noise is added to the principal components of each channel of the EEG signal. In particular, the noise added by each channel is different.
    
    .. code-block:: python

        transform = RandomPCANoise()
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

        transform = RandomPCANoise(mean=0.5, std=2.0, n_components=4)
        transform(eeg=torch.randn(1, 32, 128))['eeg'].shape
        >>> (1, 32, 128)

        transform = RandomPCANoise(p=1.0, series_dim=0)
        transform(eeg=torch.randn(128, 9, 9))['eeg'].shape
        >>> (128, 9, 9)

    Args:
        mean (float): The mean of the normal distribution of noise. (default: :obj:`0.0`)
        std (float): The standard deviation of the normal distribution of noise. (default: :obj:`0.0`)
        series_dim (int): Dimension of the time series in the input tensor. (default: :obj:`-1`)
        n_components (int): Number of components to add noise. if n_components is not set, the first two components are used to add noise.
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 mean: float = 0.0,
                 std: float = 1.0,
                 n_components: int = 2,
                 series_dim: int = -1,
                 p: float = 0.5,
                 apply_to_baseline: bool = False):
        super(RandomPCANoise,
              self).__init__(p=p, apply_to_baseline=apply_to_baseline)
        self.mean = mean
        self.std = std
        self.n_components = n_components
        self.series_dim = series_dim

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random PCA noise.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        eeg = eeg.numpy()

        assert len(eeg.shape) == 2 or len(
            eeg.shape
        ) == 3, 'Window warping is only supported on 2D arrays or 3D arrays. In 2D arrays, the last dimension represents time series. In 3D arrays, the second dimension represents time series.'

        if self.series_dim != (len(eeg.shape) - 1):
            transpose_dims = list(range(len(eeg.shape)))
            transpose_dims.pop(self.series_dim)
            transpose_dims = [*transpose_dims, self.series_dim]
            eeg = eeg.transpose(transpose_dims)

        pca = PCA(n_components=self.n_components)
        pca.fit(eeg.reshape(-1, eeg.shape[-1]))
        components = pca.components_
        variances = pca.explained_variance_ratio_

        new_eeg = np.zeros_like(eeg)
        for i, eeg_i in enumerate(eeg):
            if len(eeg.shape) == 3:
                for j, eeg_i_j in enumerate(eeg_i):
                    coeffs = np.random.normal(loc=self.mean,
                                              scale=self.std,
                                              size=pca.n_components) * variances
                    new_eeg[i][j] = eeg_i_j + (components * coeffs.reshape(
                        (pca.n_components, -1))).sum(axis=0)
            else:
                coeffs = np.random.normal(loc=self.mean,
                                          scale=self.std,
                                          size=pca.n_components) * variances
                new_eeg[i] = eeg_i + (components * coeffs.reshape(
                    (pca.n_components, -1))).sum(axis=0)

        if self.series_dim != (len(eeg.shape) - 1):
            undo_transpose_dims = [0] * len(eeg.shape)
            for i, dim in enumerate(transpose_dims):
                undo_transpose_dims[dim] = i
            new_eeg = new_eeg.transpose(undo_transpose_dims)

        return torch.from_numpy(new_eeg)

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'mean': self.mean,
                'std': self.std,
                'n_components': self.n_components,
                'series_dim': self.series_dim
            })


class RandomFlip(RandomEEGTransform):
    '''
    Applies a random transformation with a given probability to reverse the direction of the input signal in the specified dimension, commonly used for left-right and bottom-up reversal of EEG caps and reversal of timing.
    
    .. code-block:: python

        transform = RandomFlip(dim=-1)
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

        transform = RandomFlip(dim=1)
        transform(eeg=torch.randn(128, 9, 9))['eeg'].shape
        >>> (128, 9, 9)

    Args:
        dim (int): Dimension to be flipped in the input tensor. (default: :obj:`-1`)
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self, dim=-1, p: float = 0.5, apply_to_baseline: bool = False):
        super(RandomFlip, self).__init__(p=p,
                                         apply_to_baseline=apply_to_baseline)
        self.dim = dim

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random flipping.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        return torch.flip(eeg, dims=(self.dim, ))

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{'dim': self.dim})


class RandomSignFlip(RandomEEGTransform):
    '''
    Apply a random transformation such that the input signal becomes the opposite of the reversed sign with a given probability
    
    .. code-block:: python

        transform = RandomSignFlip()
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

    Args:
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self, p: float = 0.5, apply_to_baseline: bool = False):
        super(RandomSignFlip,
              self).__init__(p=p, apply_to_baseline=apply_to_baseline)

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random sign flipping.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        return -eeg


class RandomShift(RandomEEGTransform):
    '''
    Apply a shift with a specified probability, after which the specified dimension is shifted backward, and the part shifted out of the Tensor is added to the front of that dimension.
    
    .. code-block:: python

        transform = RandomShift(dim=-1, shift_min=8, shift_max=24)
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

    Args:
        shift_min (float or int): The minimum shift in the random transformation. (default: :obj:`-2.0`)
        shift_max (float or int): The maximum shift in random transformation. (default: :obj:`2.0`)
        dim (int): Dimension to be shifted in the input tensor. (default: :obj:`-1`)
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 p: float = 0.5,
                 shift_min: int = 8,
                 shift_max: int = 12,
                 dim: int = -1,
                 apply_to_baseline: bool = False):
        super(RandomShift, self).__init__(p=p,
                                          apply_to_baseline=apply_to_baseline)
        self.shift_min = shift_min
        self.shift_max = shift_max
        self.dim = dim

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random shift.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        shift = torch.randint(low=self.shift_min,
                              high=self.shift_max,
                              size=(1, ))
        return torch.roll(eeg, shifts=shift.item(), dims=self.dim)

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'shift_min': self.shift_min,
                'shift_max': self.shift_max,
                'dim': self.dim
            })


class RandomChannelShuffle(RandomEEGTransform):
    '''
    Apply a shuffle with a specified probability, after which the order of the channels is randomly shuffled.
    
    .. code-block:: python

        transform = RandomChannelShuffle()
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

    Args:
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self, p: float = 0.5, apply_to_baseline: bool = False):
        super(RandomChannelShuffle,
              self).__init__(p=p, apply_to_baseline=apply_to_baseline)

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random channel shuffle.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        index_list = np.arange(len(eeg))
        np.random.shuffle(index_list)
        return eeg[index_list]


class RandomHemisphereChannelShuffle(RandomEEGTransform):
    '''
    Apply a shuffle with a specified probability on a single hemisphere (either left or right), after which the order of the channels is randomly shuffled.
    
    .. code-block:: python

        transform = RandomChannelShuffle(location_list=M3CV_LOCATION_LIST,
                                         channel_location_dict=M3CV_CHANNEL_LOCATION_DICT)
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

    Args:
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 location_list,
                 channel_location_dict,
                 p: float = 0.5,
                 apply_to_baseline: bool = False):
        super(RandomHemisphereChannelShuffle,
              self).__init__(p=p, apply_to_baseline=apply_to_baseline)
        self.location_list = location_list
        self.channel_location_dict = channel_location_dict

        width = len(location_list[0])
        left_channel_list = []
        right_channel_list = []
        for i, (loc_y, loc_x) in enumerate(channel_location_dict.values()):
            if loc_x < width // 2:
                left_channel_list.append(i)
            if loc_y > width // 2:
                right_channel_list.append(i)
        self.left_channel_list = left_channel_list
        self.right_channel_list = right_channel_list

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random channel shuffle.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        if 0.5 < torch.rand(1):
            index_list = self.left_channel_list
        else:
            index_list = self.right_channel_list

        shuffle_index_list = np.random.permutation(index_list.copy())
        eeg[index_list] = eeg[shuffle_index_list]
        return eeg

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{
            'location_list': [...],
            'channel_location_dict': {...}
        })


class RandomFrequencyShift(RandomEEGTransform):
    '''
    Apply a frequency shift with a specified probability, after which the EEG signals of all channels are equally shifted in the frequency domain.
    
    .. code-block:: python

        transform = RandomFrequencyShift()
        transform(eeg=torch.randn(32, 128))['eeg'].shape
        >>> (32, 128)

        transform = RandomFrequencyShift(sampling_rate=128, shift_min=4.0)
        transform(eeg=torch.randn(1, 32, 128))['eeg'].shape
        >>> (1, 32, 128)

        transform = RandomFrequencyShift(p=1.0, series_dim=0)
        transform(eeg=torch.randn(128, 9, 9))['eeg'].shape
        >>> (128, 9, 9)

    Args:
        sampling_rate (int): The original sampling rate in Hz (default: :obj:`128`)
        shift_min (float or int): The minimum shift in the random transformation. (default: :obj:`-2.0`)
        shift_max (float or int): The maximum shift in random transformation. (default: :obj:`2.0`)
        series_dim (int): Dimension of the time series in the input tensor. (default: :obj:`-1`)
        p (float): Probability of applying random mask on EEG signal samples. Should be between 0.0 and 1.0, where 0.0 means no mask is applied to every sample and 1.0 means that masks are applied to every sample. (default: :obj:`0.5`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 p: float = 0.5,
                 sampling_rate: int = 128,
                 shift_min: Union[float, int] = -2.0,
                 shift_max: Union[float, int] = 2.0,
                 series_dim: int = 0,
                 apply_to_baseline: bool = False):
        super(RandomFrequencyShift,
              self).__init__(p=p, apply_to_baseline=apply_to_baseline)
        self.sampling_rate = sampling_rate
        self.shift_min = shift_min
        self.shift_max = shift_max
        self.series_dim = series_dim

    def __call__(self,
                 *args,
                 eeg: torch.Tensor,
                 baseline: Union[torch.Tensor, None] = None,
                 **kwargs) -> Dict[str, torch.Tensor]:
        r'''
        Args:
            eeg (torch.Tensor): The input EEG signal.
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch.Tensor: The output EEG signal after applying a random sampling_rate shift.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def random_apply(self, eeg: torch.Tensor, **kwargs) -> torch.Tensor:
        if self.series_dim != (len(eeg.shape) - 1):
            permute_dims = list(range(len(eeg.shape)))
            permute_dims.pop(self.series_dim)
            permute_dims = [*permute_dims, self.series_dim]
            eeg = eeg.permute(permute_dims)

        N_orig = eeg.shape[-1]
        N_padded = 2**int(np.ceil(np.log2(np.abs(N_orig))))
        t = torch.arange(N_padded) / self.sampling_rate
        padded = pad(eeg, (0, N_padded - N_orig))

        if torch.is_complex(eeg):
            raise ValueError("eeg must be real.")

        N = padded.shape[-1]
        f = fft(padded, N, dim=-1)
        h = torch.zeros_like(f)
        if N % 2 == 0:
            h[..., 0] = h[..., N // 2] = 1
            h[..., 1:N // 2] = 2
        else:
            h[..., 0] = 1
            h[..., 1:(N + 1) // 2] = 2

        analytical = ifft(f * h, dim=-1)

        shift = torch.rand(1) * (self.shift_max -
                                 self.shift_min) + self.shift_min
        shifted = analytical * torch.exp(2j * np.pi * shift * t)

        shifted = shifted[..., :N_orig].real.float()

        if self.series_dim != (len(eeg.shape) - 1):
            undo_permute_dims = [0] * len(eeg.shape)
            for i, dim in enumerate(permute_dims):
                undo_permute_dims[dim] = i
            shifted = shifted.permute(undo_permute_dims)

        return shifted

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'sampling_rate': self.sampling_rate,
                'shift_min': self.shift_min,
                'shift_max': self.shift_max,
                'series_dim': self.series_dim
            })



import numpy as np

from typing import List, Tuple, Dict


def format_region_channel_list(channel_list, region_list):
    output = []
    for region in region_list:
        region_channel_index_list = []
        for region_channel in region:
            try:
                channel_index = channel_list.index(region_channel)
            except:
                continue
            region_channel_index_list.append(channel_index)
        if len(region_channel_index_list) > 0:
            output.append(region_channel_index_list)
    return output


def format_channel_location_dict(channel_list, location_list):
    location_list = np.array(location_list)
    output = {}
    for channel in channel_list:
        if len(np.argwhere(location_list == channel)):
            location = (np.argwhere(location_list == channel)[0]).tolist()
            output[channel] = location
        else:
            output[channel] = None
    return output


def format_adj_matrix_from_adj_list(channel_list: List,
                                    adj_list: List) -> List[List]:
    node_map = {k: i for i, k in enumerate(channel_list)}
    adj_matrix = np.zeros((len(channel_list), len(channel_list)))

    for start_node_name in adj_list:
        if not start_node_name in channel_list:
            continue
        start_node_index = node_map[start_node_name]
        end_node_list = adj_list[start_node_name]

        for end_node_name in end_node_list:
            if not end_node_name in node_map:
                continue
            end_node_index = node_map[end_node_name]
            adj_matrix[start_node_index][end_node_index] = 1

    return adj_matrix.tolist()


DEFAULT_GLOBAL_CHANNEL_LIST = [('FP1', 'FP2'), ('AF3', 'AF4'), ('F5', 'F6'),
                               ('FC5', 'FC6'), ('C5', 'C6'), ('CP5', 'CP6'),
                               ('P5', 'P6'), ('PO5', 'PO6'), ('O1', 'O2')]


def format_adj_matrix_from_standard(
    channel_list: List,
    standard_channel_location_dict: Dict,
    delta: float = 0.00056,
    global_channel_list: List[Tuple[str]] = DEFAULT_GLOBAL_CHANNEL_LIST
) -> List[List]:
    r'''
    Creates an adjacency matrix based on the relative positions of electrodes in a standard system, allowing the addition of global electrode links to connect non-adjacent but symmetrical electrodes.

    - Paper: Zhong P, Wang D, Miao C. EEG-based emotion recognition using regularized graph neural networks[J]. IEEE Transactions on Affective Computing, 2020.
    - URL: https://ieeexplore.ieee.org/abstract/document/9091308
    - Related Project: https://github.com/zhongpeixiang/RGNN

    Args:
        channel_list (list): The channel name lists corresponding to the input EEG signal. If the dataset in TorchEEG is used, please refer to the CHANNEL_LIST related constants in the :obj:`torcheeg.constants` module.
        standard_channel_location_dict (dict): Electrode location information. Represented in dictionary form, where :obj:`key` corresponds to the electrode name and :obj:`value` corresponds to (x, y, z) of the electrode on the grid. please refer to STANDARD_1005_CHANNEL_LOCATION_DICT and STANDARD_1020_CHANNEL_LOCATION_DICT.
        delta (float): The calibration constant. Due to differences in electrode coordinate scales, the values in the original paper are not recommended. 0.00056 means 20% more nodes are connected to each other. (default: :obj:`0.00056`)
        global_channel_list (float): To leverage the differential asymmetry information, the authors initialize the global inter-channel relations in the adjacency matrix. (default: :obj:`[('FP1', 'FP2'), ('AF3', 'AF4'), ('F5', 'F6'), ('FC5', 'FC6'), ('C5', 'C6'), ('CP5', 'CP6'), ('P5', 'P6'), ('PO5', 'PO6'), ('O1', 'O2')]`)
    '''
    node_map = {k: i for i, k in enumerate(channel_list)}
    adj_matrix = np.zeros((len(channel_list), len(channel_list)))

    for start_node_name in channel_list:
        if not start_node_name in standard_channel_location_dict:
            continue
        for end_node_name in channel_list:
            if not end_node_name in standard_channel_location_dict:
                continue
            start_node_pos = np.array(
                standard_channel_location_dict[start_node_name])
            end_node_pos = np.array(
                standard_channel_location_dict[end_node_name])
            edge_weight = np.linalg.norm(start_node_pos - end_node_pos)
            edge_weight = min(1.0, delta / (edge_weight**2 + 1e-6))

            adj_matrix[node_map[start_node_name]][
                node_map[end_node_name]] = edge_weight

    for start_node_name, end_node_name in global_channel_list:
        if (not start_node_name in node_map) or (not end_node_name in node_map):
            continue
        adj_matrix[node_map[start_node_name]][
            node_map[end_node_name]] = adj_matrix[node_map[start_node_name]][
                node_map[end_node_name]] - 1.0

    return adj_matrix.tolist()


from typing import List

import torch.nn as nn

from .dann_like import _DANNLikeTrainer


class DANNTrainer(_DANNLikeTrainer):
    r'''
    The individual differences and nonstationary of EEG signals make it difficult for deep learning models trained on the training set of subjects to correctly classify test samples from unseen subjects, since the training set and test set come from different data distributions. Domain adaptation is used to address the problem of distribution drift between training and test sets and thus achieves good performance in subject-independent (cross-subject) scenarios. This class supports the implementation of Domain Adversarial Neural Networks (DANN) for deep domain adaptation.

    NOTE: DANN belongs to unsupervised domain adaptation methods, which only use labeled source and unlabeled target data. This means that the target dataset does not have to return labels.

    - Paper: Ganin Y, Ustinova E, Ajakan H, et al. Domain-adversarial training of neural networks[J]. The journal of machine learning research, 2016, 17(1): 2096-2030.
    - URL: https://arxiv.org/abs/1505.07818
    - Related Project: https://github.com/fungtion/DANN/blob/master/train/main.py

    .. code-block:: python

        trainer = DANNTrainer(extractor,
                              classifier,
                              domain_classifier,
                              num_classes=10,
                              devices=1,
                              accelerator='gpu')
        trainer.fit(source_loader, target_loader, val_loader)
        trainer.test(test_loader)

    Args:
        extractor (nn.Module): The feature extraction model learns the feature representation of the EEG signal by forcing the correlation matrixes of source and target data to be close.
        classifier (nn.Module): The classification model learns the classification task with the source labeled data based on the feature of the feature extraction model. The dimension of its output should be equal to the number of categories in the dataset. The output layer does not need to have a softmax activation function.
        domain_classifier (nn.Module): The classification model, learning to discriminate between the source and target domains based on the feature of the feature extraction model. The dimension of its output should be equal to the number of categories in the dataset. The output layer does not need to have a softmax activation function or a gradient reverse layer (which is already included in the implementation).
        num_classes (int, optional): The number of categories in the dataset. (default: :obj:`None`)
        lr (float): The learning rate. (default: :obj:`0.0001`)
        weight_decay (float): The weight decay. (default: :obj:`0.0`)
        weight_domain (float): The weight of the DANN loss (default: :obj:`1.0`)
        alpha_scheduler (bool): Whether to use a scheduler for the alpha of the DANN loss, growing from 0 to 1 following the schedule from the DANN paper. (default: :obj:`True`)
        lr_scheduler (bool): Whether to use a scheduler for the learning rate, as defined in the DANN paper. (default: :obj:`False`)
        warmup_epochs (int): The number of epochs for the warmup phase, during which the weight of the DANN loss is 0. (default: :obj:`0`)
        devices (int): The number of devices to use. (default: :obj:`1`)
        accelerator (str): The accelerator to use. Available options are: 'cpu', 'gpu'. (default: :obj:`"cpu"`)
        metrics (list of str): The metrics to use. Available options are: 'precision', 'recall', 'f1score', 'accuracy'. (default: :obj:`["accuracy"]`)

    .. automethod:: fit
    .. automethod:: test
    '''
    def __init__(self,
                 extractor: nn.Module,
                 classifier: nn.Module,
                 domain_classifier: nn.Module,
                 num_classes: int,
                 lr: float = 1e-4,
                 weight_decay: float = 0.0,
                 weight_domain: float = 1.0,
                 alpha_scheduler: bool = True,
                 lr_scheduler_gamma: float = 0.0,
                 lr_scheduler_decay: float = 0.75,
                 warmup_epochs: int = 0,
                 devices: int = 1,
                 accelerator: str = "cpu",
                 metrics: List[str] = ["accuracy"]):

        super(DANNTrainer, self).__init__(extractor=extractor,
                                          classifier=classifier,
                                          domain_classifier=domain_classifier,
                                          num_classes=num_classes,
                                          lr=lr,
                                          weight_decay=weight_decay,
                                          weight_domain=weight_domain,
                                          alpha_scheduler=alpha_scheduler,
                                          lr_scheduler_gamma=lr_scheduler_gamma,
                                          lr_scheduler_decay=lr_scheduler_decay,
                                          warmup_epochs=warmup_epochs,
                                          devices=devices,
                                          accelerator=accelerator,
                                          metrics=metrics)



import os
import re
from copy import copy
from typing import Tuple, Dict

import pandas as pd
from torcheeg.datasets.module.base_dataset import BaseDataset


class Subcategory:
    r'''
    A tool class for separating out subsets of specified categories, often used to extract data for a certain type of paradigm, or for a certain type of task. Each subset in the formed subset list contains only one type of data.
    
    Common usage:

    .. code-block:: python

        cv = Subcategory(split_path='./split')
        dataset = M3CVDataset(io_path=f'./m3cv',
                              root_path='./aistudio',
                              online_transform=transforms.Compose(
                                  [transforms.To2d(),
                                   transforms.ToTensor()]),
                              label_transform=transforms.Compose([
                                  transforms.Select('subject_id'),
                                  transforms.StringToInt()
                              ]))
        for subdataset in cv.split(dataset):
            loader = DataLoader(subdataset)
            ...
    
    TorchEEG supports the division of training and test sets within each subset after dividing the data into subsets. The sample code is as follows:

    .. code-block:: python

        cv = Subcategory(split_path='./split')
        dataset = M3CVDataset(io_path=f'./m3cv',
                              root_path='./aistudio',
                              online_transform=transforms.Compose(
                                  [transforms.To2d(),
                                   transforms.ToTensor()]),
                              label_transform=transforms.Compose([
                                  transforms.Select('subject_id'),
                                  transforms.StringToInt()
                              ]))
        for i, subdataset in enumerate(cv.split(dataset)):
            train_dataset, test_dataset = train_test_split(dataset=subdataset, split_path=f'./split{i}')

            train_loader = DataLoader(train_dataset)
            test_loader = DataLoader(test_dataset)
            ...
    
    For the already divided training and testing sets, TorchEEG recommends using two :obj:`Subcategory` to extract their subcategories respectively. On this basis, the :obj:`zip` function can be used to combine the subsets. It is worth noting that it is necessary to ensure that the training and test sets have the same number and variety of classes.

    .. code-block:: python

        train_cv = Subcategory(split_path='./split_train')
        train_dataset = M3CVDataset(io_path=f'./m3cv',
                              root_path='./aistudio',
                              online_transform=transforms.Compose(
                                  [transforms.To2d(),
                                   transforms.ToTensor()]),
                              label_transform=transforms.Compose([
                                  transforms.Select('subject_id'),
                                  transforms.StringToInt()
                              ]))

        val_cv = Subcategory(split_path='./split_val')
        val_dataset = M3CVDataset(io_path=f'./m3cv',
                              root_path='./aistudio',
                              subset='Calibration', num_channel=65,
                              online_transform=transforms.Compose(
                                  [transforms.To2d(),
                                   transforms.ToTensor()]),
                              label_transform=transforms.Compose([
                                  transforms.Select('subject_id'),
                                  transforms.StringToInt()
                              ]))

        for train_dataset, val_dataset in zip(train_cv.split(train_dataset), val_cv.split(val_dataset)):
            train_loader = DataLoader(train_dataset)
            val_loader = DataLoader(val_dataset)
            ...

    Args:
        criteria (str): The classification criteria according to which we extract subsets of data for the including categories. (default: :obj:`'task'`)
        split_path (str): The path to data partition information. If the path exists, read the existing partition from the path. If the path does not exist, the current division method will be saved for next use. (default: :obj:`'./split/k_fold_dataset'`)
    '''
    def __init__(self,
                 criteria: str = 'task',
                 split_path: str = './split/subcategory'):
        self.criteria = criteria
        self.split_path = split_path

    def split_info_constructor(self, info: pd.DataFrame) -> None:
        assert self.criteria in list(
            info.columns
        ), f'Unsupported criteria {self.criteria}, please select one of the following options {list(info.columns)}.'

        category_list = list(set(info[self.criteria]))

        for category in category_list:
            subset_info = info[info[self.criteria] == category]
            subset_info.to_csv(os.path.join(self.split_path, f'{category}.csv'),
                               index=False)

    @property
    def category_list(self):
        indice_files = list(os.listdir(self.split_path))

        def indice_file_to_fold_id(indice_file):
            return int(re.findall(r'(.*).csv', indice_file)[0])

        category_list = list(set(map(indice_file_to_fold_id, indice_files)))
        category_list.sort()

        return category_list

    def split(self, dataset: BaseDataset) -> Tuple[BaseDataset, BaseDataset]:
        if not os.path.exists(self.split_path):
            os.makedirs(self.split_path)
            self.split_info_constructor(dataset.info)

        category_list = self.category_list

        for category in category_list:
            subset_info = pd.read_csv(
                os.path.join(self.split_path, f'{category}.csv'))

            subset_dataset = copy(dataset)
            subset_dataset.info = subset_info

            yield subset_dataset

    @property
    def repr_body(self) -> Dict:
        return {
            'criteria': self.criteria,
            'split_path': self.split_path
        }

    def __repr__(self) -> str:
        # init info
        format_string = self.__class__.__name__ + '('
        for i, (k, v) in enumerate(self.repr_body.items()):
            # line end
            if i:
                format_string += ', '
            # str param
            if isinstance(v, str):
                format_string += f"{k}='{v}'"
            else:
                format_string += f"{k}={v}"
        format_string += ')'
        return format_string


import os
import re
from typing import Any, Callable, Dict, Tuple, Union

import numpy as np
import scipy.io as scio

from torcheeg.io import EEGSignalIO, MetaInfoIO

from ...constants.emotion_recognition.seed_iv import (
    SEED_IV_ADJACENCY_MATRIX, SEED_IV_CHANNEL_LOCATION_DICT)
from ..base_dataset import BaseDataset


class SEEDIVFeatureDataset(BaseDataset):
    r'''
    The SEED-IV dataset provided by the BCMI laboratory, which is led by Prof. Bao-Liang Lu. Since the SEED dataset provides features based on matlab, this class implements the processing of these feature files to initialize the dataset. The relevant information of the dataset is as follows:

    - Author: Zheng et al.
    - Year: 2018
    - Download URL: https://ieeexplore.ieee.org/abstract/document/8283814
    - Reference: Zheng W L, Liu W, Lu Y, et al. Emotionmeter: A multimodal framework for recognizing human emotions[J]. IEEE transactions on cybernetics, 2018, 49(3): 1110-1122.
    - Stimulus: 168 film clips.
    - Signals: Electroencephalogram (62 channels at 200Hz) and eye movement data of 15 subjects (8 females). Each subject conducts the experiments in three sessions, and each session contains 24 trials (6 per emotional category) totally 15 people x 3 sessions x 24 trials.
    - Rating: neutral (0), sad (1), fear (2), and happy (3).
    - Features: de_movingAve, de_LDS, psd_movingAve, psd_LDS, dasm_movingAve, dasm_LDS, rasm_movingAve, rasm_LDS, asm_movingAve, asm_LDS, dcau_movingAve, dcau_LDS of 4-second long windows

    In order to use this dataset, the download folder :obj:`eeg_feature_smooth` is required, containing the following folder:
    
    - 1
    - 2
    - 3

    An example dataset for CNN-based methods:

    .. code-block:: python

        dataset = SEEDIVFeatureDataset(io_path=f'./seed',
                              root_path='./eeg_feature_smooth',
                              features=['de_movingAve'],
                              offline_transform=transforms.ToGrid(SEED_CHANNEL_LOCATION_DICT),
                              online_transform=transforms.ToTensor(),
                              label_transform=transforms.Select('emotion'))
        print(dataset[0])
        # EEG signal (torch.Tensor[5, 9, 9]),
        # coresponding baseline signal (torch.Tensor[5, 9, 9]),
        # label (int)

    An example dataset for GNN-based methods:

    .. code-block:: python
    
        dataset = SEEDIVFeatureDataset(io_path=f'./seed',
                              root_path='./eeg_feature_smooth',
                              features=['de_movingAve'],
                              online_transform=ToG(SEED_ADJACENCY_MATRIX),
                              label_transform=transforms.Select('emotion'))
        print(dataset[0])
        # EEG signal (torch_geometric.data.Data),
        # coresponding baseline signal (torch_geometric.data.Data),
        # label (int)
        
    In particular, TorchEEG utilizes the producer-consumer model to allow multi-process data preprocessing. If your data preprocessing is time consuming, consider increasing :obj:`num_worker` for higher speedup. If running under Windows, please use the proper idiom in the main module:

    .. code-block:: python
    
        if __name__ == '__main__':
            dataset = SEEDIVFeatureDataset(io_path=f'./seed',
                              root_path='./eeg_feature_smooth',
                              feature=['de_movingAve'],
                              offline_transform=transforms.ToGrid(SEED_CHANNEL_LOCATION_DICT),
                              online_transform=transforms.ToTensor(),
                              label_transform=transforms.Select('emotion'),
                              num_worker=4)
            print(dataset[0])
            # EEG signal (torch_geometric.data.Data),
            # coresponding baseline signal (torch_geometric.data.Data),
            # label (int)

    Args:
        root_path (str): Downloaded data files in matlab (unzipped ExtractedFeatures.zip) formats (default: :obj:`'./ExtractedFeatures'`)
        feature (list): A list of selected feature names. The selected features corresponding to each electrode will be concatenated together. Feature names supported by the SEED dataset include de_movingAve, de_LDS, psd_movingAve, and etc. (default: :obj:`['de_movingAve']`)
        num_channel (int): Number of channels used, of which the first 62 channels are EEG signals. (default: :obj:`62`)
        online_transform (Callable, optional): The transformation of the EEG signals and baseline EEG signals. The input is a :obj:`np.ndarray`, and the ouput is used as the first and second value of each element in the dataset. (default: :obj:`None`)
        offline_transform (Callable, optional): The usage is the same as :obj:`online_transform`, but executed before generating IO intermediate results. (default: :obj:`None`)
        label_transform (Callable, optional): The transformation of the label. The input is an information dictionary, and the ouput is used as the third value of each element in the dataset. (default: :obj:`None`)
        before_trial (Callable, optional): The hook performed on the trial to which the sample belongs. It is performed before the offline transformation and thus typically used to implement context-dependent sample transformations, such as moving averages, etc. The input of this hook function is a 3D EEG signal with shape (number of windows, number of electrodes, number of features), whose ideal output shape is also (number of windows, number of electrodes, number of features).
        after_trial (Callable, optional): The hook performed on the trial to which the sample belongs. It is performed after the offline transformation and thus typically used to implement context-dependent sample transformations, such as moving averages, etc. The input and output of this hook function should be a sequence of dictionaries representing a sequence of EEG samples. Each dictionary contains two key-value pairs, indexed by :obj:`eeg` (the EEG signal matrix) and :obj:`key` (the index in the database) respectively.
        io_path (str): The path to generated unified data IO, cached as an intermediate result. (default: :obj:`./io/seed_iv_feature`)
        io_size (int): Maximum size database may grow to; used to size the memory mapping. If database grows larger than ``map_size``, an exception will be raised and the user must close and reopen. (default: :obj:`10485760`)
        io_mode (str): Storage mode of EEG signal. When io_mode is set to :obj:`lmdb`, TorchEEG provides an efficient database (LMDB) for storing EEG signals. LMDB may not perform well on limited operating systems, where a file system based EEG signal storage is also provided. When io_mode is set to :obj:`pickle`, pickle-based persistence files are used. (default: :obj:`lmdb`)
        num_worker (int): Number of subprocesses to use for data loading. 0 means that the data will be loaded in the main process. (default: :obj:`0`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
        in_memory (bool): Whether to load the entire dataset into memory. If :obj:`in_memory` is set to True, then the first time an EEG sample is read, the entire dataset is loaded into memory for subsequent retrieval. Otherwise, the dataset is stored on disk to avoid the out-of-memory problem. (default: :obj:`False`)    
    '''
    channel_location_dict = SEED_IV_CHANNEL_LOCATION_DICT
    adjacency_matrix = SEED_IV_ADJACENCY_MATRIX

    def __init__(self,
                 root_path: str = './eeg_feature_smooth',
                 feature: list = ['de_movingAve'],
                 num_channel: int = 62,
                 online_transform: Union[None, Callable] = None,
                 offline_transform: Union[None, Callable] = None,
                 label_transform: Union[None, Callable] = None,
                 before_trial: Union[None, Callable] = None,
                 after_trial: Union[Callable, None] = None,
                 after_session: Union[Callable, None] = None,
                 after_subject: Union[Callable, None] = None,
                 io_path: str = './io/seed_iv_feature',
                 io_size: int = 10485760,
                 io_mode: str = 'lmdb',
                 num_worker: int = 0,
                 verbose: bool = True,
                 in_memory: bool = False):
        # pass all arguments to super class
        params = {
            'root_path': root_path,
            'feature': feature,
            'num_channel': num_channel,
            'online_transform': online_transform,
            'offline_transform': offline_transform,
            'label_transform': label_transform,
            'before_trial': before_trial,
            'after_trial': after_trial,
            'after_session': after_session,
            'after_subject': after_subject,
            'io_path': io_path,
            'io_size': io_size,
            'io_mode': io_mode,
            'num_worker': num_worker,
            'verbose': verbose,
            'in_memory': in_memory
        }
        super().__init__(**params)
        # save all arguments to __dict__
        self.__dict__.update(params)

    @staticmethod
    def process_record(feature: list = ['de_movingAve'],
                       num_channel: int = 62,
                       offline_transform: Union[None, Callable] = None,
                       before_trial: Union[None, Callable] = None,
                       after_trial: Union[Callable, None] = None,
                       file: Any = None,
                       **kwargs):
        file_path = file  # an element from file name list

        session_id = os.path.basename(os.path.dirname(file_path))
        _, file_name = os.path.split(file_path)

        subject = int(os.path.basename(file_name).split('.')[0].split('_')
                      [0])  # subject (15)
        date = int(os.path.basename(file_name).split('.')[0].split('_')
                   [1])  # period (3)

        samples = scio.loadmat(file_path,
                               verify_compressed_data_integrity=False
                               )  # trial (15), channel(62), timestep(n*200)
        # label file
        labels = [
            [
                1, 2, 3, 0, 2, 0, 0, 1, 0, 1, 2, 1, 1, 1, 2, 3, 2, 2, 3, 3, 0,
                3, 0, 3
            ],
            [
                2, 1, 3, 0, 0, 2, 0, 2, 3, 3, 2, 3, 2, 0, 1, 1, 2, 1, 0, 3, 0,
                1, 3, 1
            ],
            [
                1, 2,
                2, 1,
                3, 3,
                3, 1,
                1, 2,
                1, 0,
                2, 3,
                3, 0,
                2, 3,
                0, 0,
                2, 0,
                1, 0
            ]
        ]  # The labels with 0, 1, 2, and 3 denote the ground truth, neutral, sad, fear, and happy emotions, respectively.
        session_labels = labels[int(session_id) - 1]
        trial_ids = [
            int(re.findall(r"de_movingAve(\d+)", key)[0])
            for key in samples.keys() if 'de_movingAve' in key
        ]

        write_pointer = 0
        # loop for each trial
        for trial_id in trial_ids:
            # extract baseline signals
            trial_samples = []
            for cur_feature in feature:
                trial_samples.append(samples[
                    cur_feature +
                    str(trial_id)])  # channel(61), timestep(n), bands(5)
            trial_samples = np.concatenate(
                trial_samples,
                axis=-1)[:
                         num_channel]  # channel(61), timestep(n), features(5*k)
            trial_samples = trial_samples.transpose((1, 0, 2))
            # timestep(n), channel(61), features(5*k)

            if before_trial:
                trial_samples = before_trial(trial_samples)

            # record the common meta info
            trial_meta_info = {
                'subject_id': subject,
                'trial_id': trial_id,
                'session_id': session_id,
                'emotion': int(session_labels[trial_id - 1]),
                'date': date
            }

            trial_queue = []
            for i, clip_sample in enumerate(trial_samples):
                t_eeg = clip_sample
                if not offline_transform is None:
                    t_eeg = offline_transform(eeg=clip_sample)['eeg']

                clip_id = f'{file_name}_{write_pointer}'
                write_pointer += 1

                # record meta info for each signal
                record_info = {
                    'start_at': i * 400,
                    'end_at': (i + 1) *
                    400,  # The size of the sliding time windows for feature extraction is 4 seconds.
                    'clip_id': clip_id
                }
                record_info.update(trial_meta_info)
                yield {'eeg': t_eeg, 'key': clip_id, 'info': record_info}
                
    def set_records(self, root_path: str = './eeg_feature_smooth', **kwargs):

        session_list = ['1', '2', '3']
        file_path_list = []
        for session in session_list:
            session_root_path = os.path.join(root_path, session)
            for file_name in os.listdir(session_root_path):
                file_path_list.append(os.path.join(session_root_path,
                                                   file_name))

        return file_path_list

    def __getitem__(self, index: int) -> Tuple[any, any, int, int, int]:
        info = self.read_info(index)

        eeg_index = str(info['clip_id'])
        eeg_record = str(info['_record_id'])
        eeg = self.read_eeg(eeg_record, eeg_index)

        signal = eeg
        label = info

        if self.online_transform:
            signal = self.online_transform(eeg=eeg)['eeg']

        if self.label_transform:
            label = self.label_transform(y=info)['y']

        return signal, label

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'root_path': self.root_path,
                'feature': self.feature,
                'num_channel': self.num_channel,
                'online_transform': self.online_transform,
                'offline_transform': self.offline_transform,
                'label_transform': self.label_transform,
                'before_trial': self.before_trial,
                'after_trial': self.after_trial,
                'num_worker': self.num_worker,
                'verbose': self.verbose,
                'io_size': self.io_size
            })



import torch
import torch.nn as nn
import torch.nn.functional as F


class GRU(nn.Module):
    r'''
    A simple but effective gate recurrent unit (GRU) network structure from the book of Zhang et al. For more details, please refer to the following information.

    - Book: Zhang X, Yao L. Deep Learning for EEG-Based Brain-Computer Interfaces: Representations, Algorithms and Applications[M]. 2021.
    - URL: https://www.worldscientific.com/worldscibooks/10.1142/q0282#t=aboutBook
    - Related Project: https://github.com/xiangzhang1015/Deep-Learning-for-BCI/blob/master/pythonscripts/4-1-2_GRU.py

    Below is a recommended suite for use in emotion recognition tasks:

    .. code-block:: python

        dataset = DEAPDataset(io_path=f'./deap',
                    root_path='./data_preprocessed_python',
                    online_transform=transforms.ToTensor(),
                    label_transform=transforms.Compose([
                        transforms.Select('valence'),
                        transforms.Binary(5.0),
                    ]))
        model = GRU(num_electrodes=32, hid_channels=64, num_classes=2)

    Args:
        num_electrodes (int): The number of electrodes, i.e., :math:`C` in the paper. (default: :obj:`32`)
        hid_channels (int): The number of hidden nodes in the GRU layers and the fully connected layer. (default: :obj:`64`)
        num_classes (int): The number of classes to predict. (default: :obj:`2`)
    '''
    def __init__(self,
                 num_electrodes: int = 32,
                 hid_channels: int = 64,
                 num_classes: int = 2):
        super(GRU, self).__init__()

        self.num_electrodes = num_electrodes
        self.hid_channels = hid_channels
        self.num_classes = num_classes

        self.gru_layer = nn.GRU(
            input_size=num_electrodes,
            hidden_size=hid_channels,
            num_layers=2,
            bias=True,
            batch_first=True,
        )

        self.out = nn.Linear(hid_channels, num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        r'''
        Args:
            x (torch.Tensor): EEG signal representation, the ideal input shape is :obj:`[n, 32, 128]`. Here, :obj:`n` corresponds to the batch size, :obj:`32` corresponds to :obj:`num_electrodes`, and :obj:`128` corresponds to the number of data points included in the input EEG chunk.

        Returns:
            torch.Tensor[number of sample, number of classes]: the predicted probability that the samples belong to the classes.
        '''
        x = x.permute(0, 2, 1)

        r_out, (_, _) = self.gru_layer(x, None)
        r_out = F.dropout(r_out, 0.3)
        x = self.out(r_out[:, -1, :])  # choose r_out at the last time step
        return x


from ..region_1020 import (FRONTAL_REGION_LIST, GENERAL_REGION_LIST,
                           HEMISPHERE_REGION_LIST)
from ..standard_1005 import STANDARD_1005_CHANNEL_LOCATION_DICT
from ..utils import (format_adj_matrix_from_adj_list,
                     format_adj_matrix_from_standard,
                     format_channel_location_dict, format_region_channel_list)

TUHTUEG_CHANNEL_LIST = [
    'FP1', 'FP2', 'F3', 'F4', 'C3', 'C4', 'P3', 'P4', 'O1', 'O2', 'F7', 'F8',
    'T3', 'T4', 'T5', 'T6', 'T1', 'T2', 'FZ', 'CZ', 'PZ'
]

TUHTUEG_LOCATION_LIST = [
    ['-',  '-', '-',  'FP1', '-',  'FP2', '-',  '-', '-'],
    ['-',  '-', '-',  '-',  '-',  '-',   '-',  '-', '-'],
    ['F7', '-', 'F3', '-',  'FZ', '-',   'F4', '-', 'F8'],
    ['T1', '-', '-',  '-',  '-',  '-',   '-',  '-', 'T2'],
    ['T3', '-', 'C3', '-',  'CZ', '-',   'C4', '-', 'T4'],
    ['-',  '-', '-',  '-',  '-',  '-',   '-',  '-', '-'],
    ['T5', '-', 'P3', '-',  'PZ', '-',   'P4', '-', 'T6'],
    ['-',  '-', '-',  '-',  '-',  '-',   '-',  '-', '-'],
    ['-',  '-', '-',  'O1', '-',  'O2',  '-',  '-', '-']
]

TUHTUEG_CHANNEL_LOCATION_DICT = format_channel_location_dict(
    TUHTUEG_CHANNEL_LIST, TUHTUEG_LOCATION_LIST)

TUHTUEG_ADJACENCY_LIST = {
    'FP1': ['FP2', 'F3', 'FZ'],
    'FP2': ['FP1', 'FZ', 'F4'],
    'F7': ['T1', 'T3'],
    'F3': ['FP1', 'FZ', 'C3', 'CZ'],
    'FZ': ['FP1', 'FP2', 'F3', 'F4', 'C3', 'CZ', 'C4'],
    'F4': ['FP2', 'FZ', 'CZ', 'C4'],
    'F8': ['T2', 'T4'],
    'T1': ['F7', 'T3'],
    'T2': ['F8', 'T4'],
    'T3': ['F7', 'T1', 'T5'],
    'C3': ['F3', 'FZ', 'CZ', 'P3', 'PZ'],
    'CZ': ['F3', 'FZ', 'F4', 'C3', 'C4', 'P3', 'PZ', 'P4'],
    'C4': ['FZ', 'F4', 'CZ', 'PZ', 'P4'],
    'T4': ['F8', 'T2', 'T6'],
    'T5': ['T3'],
    'P3': ['C3', 'CZ', 'PZ', 'O1'],
    'PZ': ['C3', 'CZ', 'C4', 'P3', 'P4', 'O1', 'O2'],
    'P4': ['CZ', 'C4', 'PZ', 'O2'],
    'T6': ['T4'],
    'O1': ['P3', 'PZ', 'O2'],
    'O2': ['PZ', 'P4', 'O1']
}

TUHTUEG_ADJACENCY_MATRIX = format_adj_matrix_from_adj_list(
    TUHTUEG_CHANNEL_LIST, TUHTUEG_ADJACENCY_LIST)

TUHTUEG_STANDARD_ADJACENCY_MATRIX = format_adj_matrix_from_standard(
    TUHTUEG_CHANNEL_LIST, STANDARD_1005_CHANNEL_LOCATION_DICT)

TUHTUEG_GENERAL_REGION_LIST = format_region_channel_list(
    TUHTUEG_CHANNEL_LIST, GENERAL_REGION_LIST)
TUHTUEG_FRONTAL_REGION_LIST = format_region_channel_list(
    TUHTUEG_CHANNEL_LIST, FRONTAL_REGION_LIST)
TUHTUEG_HEMISPHERE_REGION_LIST = format_region_channel_list(
    TUHTUEG_CHANNEL_LIST, HEMISPHERE_REGION_LIST)